@@ -1,11 +1,9 @@
-import asyncio
 import datetime
 import os
 import random
 import string
 import uuid
 
-import aiofiles
 import jwt
 from captcha.image import ImageCaptcha
 from jwt import DecodeError
@@ -15,8 +13,9 @@ from sanic.response import HTTPResponse, file
 from tortoise import fields, Model
 from tortoise.exceptions import DoesNotExist
 
+from sanic_security.configuration import config as security_config
 from sanic_security.exceptions import *
-from sanic_security.utils import get_ip, dir_exists, config
+from sanic_security.utils import get_ip, dir_exists
 
 """
 Copyright (C) 2021 Aidan Stewart
@@ -98,7 +97,7 @@ class Account(BaseModel):
         username (str): Public identifier.
         email (str): Private identifier and can be used for verification.
         phone (str): Mobile phone number with country code included and can be used for verification. May be null or empty.
-        password (bytes): Password of account for protection. Must be hashed via bcrypt.
+        password (str): Password of account for protection. Must be hashed via Argon.
         disabled (bool): Renders the account unusable but available.
         verified (bool): Renders the account unusable until verified via two-step verification or other method.
     """
@@ -106,7 +105,7 @@ class Account(BaseModel):
     username = fields.CharField(max_length=32)
     email = fields.CharField(unique=True, max_length=255)
     phone = fields.CharField(unique=True, max_length=14, null=True)
-    password = fields.BinaryField()
+    password = fields.CharField(max_length=255)
     disabled = fields.BooleanField(default=False)
     verified = fields.BooleanField(default=False)
 
@@ -198,7 +197,6 @@ class Session(BaseModel):
         expiration_date (datetime): Time the session expires and can no longer be used.
         valid (bool): Renders the session accessible but unusable.
         ip (str): IP address of client creating session.
-        cache_path (str): Session cache path.
     """
 
     expiration_date = fields.DatetimeField(null=True)
@@ -242,13 +240,12 @@ class Session(BaseModel):
             )
             raise SessionError("Unrecognised location.", 401)
 
-    def encode(self, response: HTTPResponse, tag: str = "sec"):
+    def encode(self, response: HTTPResponse):
         """
         Transforms session into jwt and then is stored in a cookie.
 
         Args:
             response (HTTPResponse): Sanic response used to store JWT into a cookie on the client.
-            tag (str): Identifier applied to encoded session cookie.
         """
         payload = {
             "date_created": str(self.date_created),
@@ -256,24 +253,25 @@ class Session(BaseModel):
             "uid": str(self.uid),
             "ip": self.ip,
         }
-        cookie = f"{tag}_{self.__class__.__name__}"
+        cookie = f"{security_config.SESSION_PREFIX}_{self.__class__.__name__.lower()[:4]}_session"
         response.cookies[cookie] = jwt.encode(
-            payload, config["SECURITY"]["secret"], "HS256"
-        )
-        response.cookies[cookie]["httponly"] = True
-        response.cookies[cookie]["samesite"] = config["SECURITY"]["session_samesite"]
-        response.cookies[cookie]["secure"] = config["SECURITY"].getboolean(
-            "session_secure"
+            payload, security_config.SECRET, security_config.SESSION_ENCODING_ALGORITHM
         )
+        response.cookies[cookie]["httponly"] = security_config.SESSION_HTTPONLY
+        response.cookies[cookie]["samesite"] = security_config.SESSION_SAMESITE
+        response.cookies[cookie]["secure"] = security_config.SESSION_SECURE
+        if security_config.SESSION_EXPIRES_ON_CLIENT and self.expiration_date:
+            response.cookies[cookie]["expires"] = self.expiration_date
+        if security_config.SESSION_DOMAIN:
+            response.cookies[cookie]["domain"] = security_config.SESSION_DOMAIN
 
     @classmethod
-    def decode_raw(cls, request: Request, tag: str = "sec") -> dict:
+    def decode_raw(cls, request: Request) -> dict:
         """
         Decodes JWT token from client cookie into a python dict.
 
         Args:
             request (Request): Sanic request parameter.
-            tag (str): Identifier found applied to encoded session cookie.
 
         Returns:
             session_dict
@@ -281,23 +279,28 @@ class Session(BaseModel):
         Raises:
             SessionError
         """
-        cookie = request.cookies.get(f"{tag}_{cls.__name__}")
+        cookie = request.cookies.get(
+            f"{security_config.SESSION_PREFIX}_{cls.__name__.lower()[:4]}_session"
+        )
         try:
             if not cookie:
                 raise SessionError(f"No session provided by client.", 400)
             else:
-                return jwt.decode(cookie, config["SECURITY"]["secret"], "HS256")
+                return jwt.decode(
+                    cookie,
+                    security_config.SECRET,
+                    security_config.SESSION_ENCODING_ALGORITHM,
+                )
         except DecodeError as e:
             raise SessionError(str(e), 400)
 
     @classmethod
-    async def decode(cls, request: Request, tag: str = "sec"):
+    async def decode(cls, request: Request):
         """
         Decodes JWT token from client cookie to a Sanic Security session.
 
         Args:
             request (Request): Sanic request parameter.
-            tag (str): Identifier found applied to encoded session cookie.
 
         Returns:
             session
@@ -306,7 +309,7 @@ class Session(BaseModel):
             NotFoundError
             SessionError
         """
-        decoded = cls.decode_raw(request, tag)
+        decoded = cls.decode_raw(request)
         try:
             decoded_session = (
                 await cls.filter(uid=decoded["uid"]).prefetch_related("account").get()
@@ -326,21 +329,22 @@ class VerificationSession(Session):
     Attributes:
         attempts (int): The amount of incorrect times a user entered a code not equal to this verification sessions code.
         code (str): Used as a secret key that would be sent via email, text, etc to complete the verification challenge.
+        cache (str): Session cache path.
     """
 
     attempts = fields.IntField(default=0)
     code = fields.CharField(max_length=10, null=True)
-    cache_path = config["SECURITY"]["cache_path"]
+    cache = security_config.CACHE
 
     @classmethod
-    async def _initialize_cache(cls):
+    def _initialize_cache(cls):
         """
         Creates verification session cache and generates required files.
         """
         raise NotImplementedError()
 
     @classmethod
-    async def get_random_code(cls) -> str:
+    def get_random_code(cls) -> str:
         """
         Retrieves a random cached verification session code.
         """
@@ -385,26 +389,21 @@ class TwoStepSession(VerificationSession):
     """
 
     @classmethod
-    async def _initialize_cache(cls):
-        if not dir_exists(f"{cls.cache_path}/verification"):
-            async with aiofiles.open(
-                f"{cls.cache_path}/verification/codes.txt", mode="w"
-            ) as f:
+    def _initialize_cache(cls):
+        if not dir_exists(f"{cls.cache}/verification"):
+            with open(f"{cls.cache}/verification/codes.txt", "w") as f:
                 for i in range(100):
                     code = "".join(
                         random.choices(string.ascii_letters + string.digits, k=10)
                     )
-                    await f.write(code + " ")
+                    f.write(f"{code} ")
             logger.info("Two-step session cache initialised")
 
     @classmethod
-    async def get_random_code(cls):
-        await cls._initialize_cache()
-        async with aiofiles.open(
-            f"{cls.cache_path}/verification/codes.txt", mode="r"
-        ) as f:
-            codes = await f.read()
-            return random.choice(codes.split())
+    def get_random_code(cls):
+        cls._initialize_cache()
+        with open(f"{cls.cache}/verification/codes.txt", "r") as f:
+            return random.choice(f.read().split())
 
     class Meta:
         table = "two_step_session"
@@ -416,26 +415,20 @@ class CaptchaSession(VerificationSession):
     """
 
     @classmethod
-    async def _initialize_cache(cls):
-        if not dir_exists(f"{cls.cache_path}/captcha"):
-            loop = asyncio.get_running_loop()
-            image = ImageCaptcha(190, 90, fonts=[config["SECURITY"]["captcha_font"]])
+    def _initialize_cache(cls):
+        if not dir_exists(f"{cls.cache}/captcha"):
+            image = ImageCaptcha(190, 90, fonts=[security_config.CAPTCHA_FONT])
             for i in range(100):
                 code = "".join(
                     random.choices("123456789qQeErRtTyYiIaAdDfFgGhHlLbBnN", k=6)
                 )
-                await loop.run_in_executor(
-                    None,
-                    image.write,
-                    code,
-                    f"{cls.cache_path}/captcha/{code}.png",
-                )
+                image.write(code, f"{cls.cache}/captcha/{code}.png")
             logger.info("Captcha session cache initialised")
 
     @classmethod
-    async def get_random_code(cls):
-        await cls._initialize_cache()
-        return random.choice(os.listdir(f"{cls.cache_path}/captcha")).split(".")[0]
+    def get_random_code(cls):
+        cls._initialize_cache()
+        return random.choice(os.listdir(f"{cls.cache}/captcha")).split(".")[0]
 
     async def get_image(self) -> HTTPResponse:
         """
@@ -444,7 +437,7 @@ class CaptchaSession(VerificationSession):
         Returns:
             captcha_image
         """
-        return await file(f"{self.cache_path}/captcha/{self.code}.png")
+        return await file(f"{self.cache}/captcha/{self.code}.png")
 
     class Meta:
         table = "captcha_session"
@@ -491,18 +484,24 @@ class SessionFactory:
             return await CaptchaSession.create(
                 **kwargs,
                 ip=get_ip(request),
-                code=await CaptchaSession.get_random_code(),
+                code=CaptchaSession.get_random_code(),
                 expiration_date=datetime.datetime.utcnow()
-                + datetime.timedelta(minutes=1),
+                + datetime.timedelta(seconds=security_config.CAPTCHA_SESSION_EXPIRATION)
+                if security_config.CAPTCHA_SESSION_EXPIRATION != 0
+                else None,
             )
         elif session_type == "two-step":
             return await TwoStepSession.create(
                 **kwargs,
-                code=await TwoStepSession.get_random_code(),
+                code=TwoStepSession.get_random_code(),
                 ip=get_ip(request),
                 account=account,
                 expiration_date=datetime.datetime.utcnow()
-                + datetime.timedelta(minutes=5),
+                + datetime.timedelta(
+                    seconds=security_config.TWO_STEP_SESSION_EXPIRATION
+                )
+                if security_config.TWO_STEP_SESSION_EXPIRATION != 0
+                else None,
             )
         elif session_type == "authentication":
             return await AuthenticationSession.create(
@@ -510,7 +509,11 @@ class SessionFactory:
                 account=account,
                 ip=get_ip(request),
                 expiration_date=datetime.datetime.utcnow()
-                + datetime.timedelta(days=30),
+                + datetime.timedelta(
+                    seconds=security_config.AUTHENTICATION_SESSION_EXPIRATION
+                )
+                if security_config.AUTHENTICATION_SESSION_EXPIRATION != 0
+                else None,
             )
         else:
             raise ValueError("Invalid session type.")
